from __future__ import annotations

from typing import Optional, Union

import pandas as pd
from matplotlib import pyplot as plt

from pltflow.utils.utils import load_styles


class plot:

    """
    Generic class to genererate a plt graph
    """

    def __init__(self, df: pd.DataFrame, x: str, y: str, style: str = "base", mode: str = "both") -> None:

        self.STYLES = load_styles()

        plt.rcParams.update(plt.rcParamsDefault)

        self.df = df
        self.x = x
        self.y = y
        self._set_mode(mode)
        self._set_style(style)
        self.set_xlabel(self.x)
        self.set_ylabel(self.y)
        self.set_figsize()
        self.z = ""  # type: str
        self.main_categories = []  # type: list
        self.notes = []  # type: list

    def color_by(self, column_name: str) -> plot:
        """
        Set the color for the plot.

        """

        if column_name not in self.df.columns:
            raise ValueError(f"{column_name} not in dataframe")

        self.z = column_name

        self.styleParams["color_by"] = column_name

        if self.main_categories == []:
            self.main_categories = list(self.df[column_name].unique())

        available_values = list(
            self.df.loc[self.df[self.z].isin(self.main_categories), self.z].value_counts().index
        )

        if len(self.main_categories) > 20:
            print(
                f"Posible categories to .focus_on()  {available_values[:20]} and {len(self.main_categories) - 20} more"
            )
        else:
            print(f"Possible categories to .focus_on() {available_values}")

        return self

    def _set_mode(self, mode: str) -> None:

        if mode in ["both", "scatter", "line"]:
            self.mode = mode
        else:
            raise ValueError("mode must be either 'both', 'scatter' or 'line'")

    def focus_on(self, category: Union[str, list]) -> plot:
        """
        Set the main category for the plot.

        """

        if self.z == "":
            raise ValueError(
                """
                No column to split on selected (z)
                Please select a column with .color_by("column_name ")
                """
            )

        all_cats_available = list(self.df[self.z].unique())

        if self.main_categories == all_cats_available:
            self.main_categories = []

        if isinstance(category, str):

            if category not in all_cats_available:
                raise ValueError(
                    f"{category} is not included on the main categories avalable in the {self.z} column"
                )
            self.main_categories.append(category)

        elif isinstance(category, list):

            invalid_categories = []
            for requested_cat in category:
                if requested_cat not in all_cats_available:
                    invalid_categories.append(requested_cat)

            if len(invalid_categories) > 0:
                raise ValueError(
                    f"{invalid_categories} does not form part of the available categories of  the column {self.x}"
                )

            self.main_categories.append(category)

        self.main_categories = list(pd.Series(self.main_categories).unique())

        return self

    def _set_style(self, style: str) -> plot:

        if style in self.STYLES:
            self.style = self.STYLES[style]
            self.styleParams = self.STYLES[style]["styleParams"]
            self.rcParams = self.STYLES[style]["rcParams"]
            self.colors = self.STYLES[style]["colors"]

        else:
            raise ValueError(f"{style} is not a valid style")

        return self

    def set_yticks(self, positions: list, **kwargs: dict) -> plot:
        self.styleParams["yticks"] = {**self.styleParams["yticks"], **{"ticks": positions}, **kwargs}
        return self

    def set_xticks(self, positions: list, **kwargs: dict) -> plot:
        self.styleParams["xticks"] = {**self.styleParams["xticks"], **{"ticks": positions}, **kwargs}
        return self

    def set_ylabel(self, label: str, **kwargs: dict) -> plot:
        self.styleParams["ylabel"] = {**self.styleParams["ylabel"], **{"ylabel": label}, **kwargs}
        return self

    def set_xlabel(self, label: str, **kwargs: dict) -> plot:
        self.styleParams["xlabel"] = {**self.styleParams["xlabel"], **{"xlabel": label}, **kwargs}
        return self

    def set_title(self, title: str, **kwargs: dict) -> plot:

        self.styleParams["title"] = {**self.styleParams["title"], **{"text": title}, **kwargs}

        return self

    def set_subtitle(self, subtitle: str) -> plot:
        self.styleParams["subtitle"]["text"] = subtitle

        return self

    def set_figsize(self, w: int = 9, h: int = 4) -> plot:
        if w > 0 and h > 0:
            self.rcParams["figure.figsize"] = (w, h)

        else:
            raise ValueError("w and h must be greater than 0")
        return self

    def _set_styleParams(self, params: dict) -> None:
        """
        Set parameters for the plot.

        """
        for key, value in params.items():
            if key in self.styleParams:
                self.styleParams[key] = value

    def add_note(
        self,
        text: str,
        x: float,
        y: Optional[float] = None,
        category: Optional[str] = None,
        **kwargs: dict,
    ) -> plot:

        if text is None:
            raise ValueError("An input text is required")

        if x is not None and y is not None:
            self.notes.append(
                [
                    {
                        "text": text,
                        "xy": (x, y),
                        "xytext": (x + 5, y + 10),
                        "arrowprops": {"arrowstyle": "->", "connectionstyle": "angle3,angleA=0,angleB=-90"},
                        "size": 15,
                    },
                    kwargs,
                ]
            )
        if category is not None and category not in self.main_categories:
            raise ValueError(
                f"{category} not in main categories. Available categories f{self.main_categories}"
            )

        if category is not None and self.z == "":
            raise ValueError(
                """
                It is not possible to specify a category on a single colored plot.
                Use color_by() to be able to specify a category
                """
            )

        if y is None:

            self._check_is_numeric(self.x)

            if category is None and len(self.main_categories) > 1:
                raise ValueError(
                    f"""
                Error: Ambiguity where to place the note You need to specify a category.
                Available categories {self.main_categories}"
                """
                )

            if category is None and self.z == "":
                selection = self.df[self.x]
                index_closest_x = abs(selection.sort_values() - x).sort_values().index[0]

            elif category is not None:
                selection = self.df[self.df[self.z] == category][self.x]
                index_closest_x = abs(selection.sort_values() - x).sort_values().index[0]

            closest_x = selection.loc[index_closest_x]
            y = self.df[self.df[self.x] == closest_x][self.y].iat[0]

            self.notes.append(
                [
                    {
                        "text": text,
                        "xy": (closest_x, y),
                        "xytext": (closest_x + 5, y + 10),
                        "arrowprops": {"arrowstyle": "->", "connectionstyle": "angle3,angleA=0,angleB=-90"},
                        "size": 15,
                    },
                    kwargs,
                ]
            )

        return self

    def _check_is_numeric(self, column: str) -> None:
        try:
            self.df[column].astype("float64")
        except ValueError:
            print(
                f"""
                {column} is not a numerical python type.
                Notes based on only one axis can only be perfom on numerical columns")
                """
            )

    def show(self) -> None:

        plt.rcParams.update(self.rcParams)

        if self.z != "":
            categories = self.df[self.z].unique().tolist()
            if len(categories) == 0:
                raise ValueError("No categories found: Length of categories is 0")
        else:
            categories = []

        if len(categories) <= 1:
            color = self.colors["1cat"]
            plt.plot(self.df[self.x], self.df[self.y], color=color, **self.style["line_style"])
            plt.scatter(
                self.df[self.x], self.df[self.y], color=color, **self.style["scatter_style"], marker="s"
            )

        else:
            colors = self.colors["ncats"]

            for category in categories:

                if category not in self.main_categories:

                    x_axis = self.df[self.x][self.df[self.z] == category]
                    y_axis = self.df[self.y][self.df[self.z] == category]

                    if self.mode in ["line", "both"]:

                        plt.plot(
                            x_axis,
                            y_axis,
                            color=self.colors["grayed"],
                            **self.style["lineshadow_style"],
                        )

                    else:

                        plt.scatter(
                            x_axis,
                            y_axis,
                            color=self.colors["grayed"],
                            **self.style["scattershadow_style"],
                        )

            for i, category in enumerate(self.main_categories):

                if len(self.main_categories) == 1:
                    color = self.colors["1cat"]
                else:
                    color = colors[i % len(colors)]

                x_axis = self.df[self.x][self.df[self.z] == category]
                y_axis = self.df[self.y][self.df[self.z] == category]

                if self.mode in ["line", "both"]:
                    plt.plot(x_axis, y_axis, color=color, **self.style["line_style"])
                if self.mode in ["scatter", "both"]:
                    plt.scatter(x_axis, y_axis, color=color, **self.style["scatter_style"])

        plt.ylabel(**self.styleParams["ylabel"])
        plt.xlabel(**self.styleParams["xlabel"])

        # plt.title(**PARAMS["title"])

        # Dirty Trick to extend the plot to the right in Jupyter notebooks
        plt.text(1, 1.09, "t", transform=plt.gcf().transFigure, color=self.rcParams["figure.facecolor"])
        plt.text(-0.05, -0.1, "t", transform=plt.gcf().transFigure, color=self.rcParams["figure.facecolor"])

        if "text" in self.styleParams["title"]:
            plt.annotate(**self.styleParams["title"])

        if "text" in self.styleParams["subtitle"]:
            plt.annotate(**self.styleParams["subtitle"])

        plt.xticks(**self.styleParams["xticks"])
        plt.yticks(**self.styleParams["yticks"])

        plt.show()



from pltflow.utils.colors import plt_tab_colors

MAIN_FONT = "Roboto mono for powerline"

style = {
    "rcParams": {
        # Background color
        "axes.facecolor": "white",
        "figure.facecolor": "white",
        # Labels settings
        "axes.labelcolor": "black",
        "axes.labelsize": 20,
        "axes.labelpad": 23,
        # XY line axis color
        "axes.edgecolor": "gray",
        # XY line width
        "axes.linewidth": 1,
        # XY line remove top/right
        "axes.spines.right": False,
        "axes.spines.top": False,
        # Change xy axis color
        "xtick.color": "black",
        "ytick.color": "black",
        "xtick.major.size": 6,
        "xtick.major.width": 1,
        "ytick.major.size": 7.2,
        "ytick.major.width": 1,
        # Y grid
        "grid.color": "lightgray",
        "grid.alpha": 0.3,
        "axes.grid": False,
        # Savefig properties
        "savefig.facecolor": "white",
        "savefig.edgecolor": "white",
    },
    "scatter_style": {
        "s": 40,
    },
    "line_style": {
        "linewidth": 1.5,
    },
    "lineshadow_style": {
        "linewidth": 1,
    },
    "scattershadow_style": {
        "s": 40,
    },
    "styleParams": {
        "xticks": {"fontsize": 11, "fontname": MAIN_FONT},
        "yticks": {"fontsize": 11, "fontname": MAIN_FONT},
        "ylabel": {
            "fontsize": 13,
            "fontname": MAIN_FONT,
            "fontweight": "bold",
            "labelpad": 14,
        },
        "xlabel": {
            "fontsize": 13,
            "fontname": MAIN_FONT,
            "labelpad": 14,
            "fontweight": "bold",
        },
        "title": {
            "fontsize": 16,
            "fontweight": "bold",
            "xy": (0.00, 1.16),
            "xycoords": "axes fraction",
            "fontname": MAIN_FONT,
        },
        "subtitle": {
            "fontsize": 12,
            "color": "#696969",
            "xy": (0.00, 1.09),
            "xycoords": "axes fraction",
            "fontname": MAIN_FONT,
        },
    },
    "colors": {"1cat": "tab:blue", "ncats": plt_tab_colors, "grayed": "lightgray"},
}



import pandas as pd


def ratio_per_category(df: pd.DataFrame, primary: str, secondary: str) -> pd.DataFrame:

    return (
        df.groupby([primary, secondary])
        .agg({secondary: "count"})
        .groupby(level=0)
        .apply(lambda x: x * 100 / x.sum())
        .rename({secondary: f"perc_{secondary}"}, axis=1)
        .reset_index()
    )



import importlib
import pkgutil

import pltflow.styles as defined_styles


def load_styles() -> dict:

    styles = [x.name for x in pkgutil.iter_modules(defined_styles.__path__)]  # type: ignore

    STYLES = {}

    for style in styles:

        STYLES[style] = importlib.import_module(f"pltflow.styles.{style}").style  # type: ignore

    return STYLES



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from skeleton.wrangling.grouping import ratio_per_category


def plot_percentage_per_subcategory(
    df: pd.DataFrame, primary: str, secondary: str
) -> Tuple[Figure, plt.Axes]:
    """
    Parameters
    ----------
    df : A pivoted dataframe with primary category as row index
         and secondary category as row columns
    primary: name of the main category in the dataframe
    secondary: name of the category to aggregate in the dataframe
    """

    per_cat_perc = ratio_per_category(df, primary, secondary)  # .sort_values(by=secondary, key=sort_order)

    pivot_per_cat_perc = (
        pd.pivot_table(
            per_cat_perc, values=f"perc_{secondary}", index=primary, columns=secondary, aggfunc=np.sum
        )
        .round(1)
        .fillna(0)
    )

    category_names = pivot_per_cat_perc.columns
    labels = list(pivot_per_cat_perc.index)
    data = pivot_per_cat_perc.values
    data_cum = data.cumsum(axis=1)
    category_colors = plt.get_cmap("RdYlGn")(np.linspace(0.15, 0.85, data.shape[1]))

    fig, ax = plt.subplots(figsize=(20, len(labels) * 2))
    ax.invert_yaxis()
    ax.xaxis.set_visible(False)
    ax.set_xlim(0, np.sum(data, axis=1).max())

    for i, (colname, color) in enumerate(zip(category_names, category_colors)):
        widths = data[:, i]
        starts = data_cum[:, i] - widths
        ax.barh(labels, widths, left=starts, height=0.8, label=colname, color=color)
        xcenters = starts + widths / 2

        r, g, b, _ = color
        text_color = "white" if r * g * b < 0.5 else "darkgrey"
        for y, (x, c) in enumerate(zip(xcenters, widths)):
            if c > 3:
                ax.text(
                    x, y, str(c) + "%", ha="center", va="center", color=text_color, size=18, weight="bold"
                )

    ax.tick_params(axis="both", labelsize=22, colors="dimgray")

    for tick in ax.get_yticklabels():
        tick.set_fontname("Arial")
        tick.set_fontweight("bold")

    # Remove contour
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["top"].set_visible(False)

    ax.legend(
        ncol=len(category_names),
        bbox_to_anchor=(0, 1),
        loc="lower left",
        fontsize="15",
        title=secondary,
        title_fontsize=16,
    )

    return fig, ax


def _sorter(df: pd.DataFrame, primary: str, secondary: str) -> pd.Series:
    """Sort function"""
    cats = list(df.groupby(primary)[secondary].mean().sort_values().index)

    correspondence = {cats: order for order, cats in enumerate(cats)}
    return df[secondary].map(correspondence)



from pltflow.utils.colors import plt_tab_colors

style = {
    "rcParams": {
        # Background color
        "axes.facecolor": "#3a4849",
        "figure.facecolor": "#3a4849",
        # Labels settings
        "axes.labelcolor": "white",
        "axes.labelsize": 20,
        "axes.labelpad": 25,
        # XY line axis color
        "axes.edgecolor": "#62737c",
        # XY line width
        "axes.linewidth": 2,
        # XY line remove top/right
        "axes.spines.right": False,
        "axes.spines.top": False,
        # Change xy axis color
        "xtick.color": "white",
        "ytick.color": "white",
        "xtick.major.size": 0,
        "ytick.major.size": 0,  # Y grid
        "grid.color": "lightgray",
        "grid.alpha": 0.4,
        "axes.grid": False,
        "axes.grid.axis": "y",
        # Savefig properties
        "savefig.facecolor": "black",
        "savefig.edgecolor": "black",
    },
    "scatter_style": {
        "s": 30,
        "marker": "s",
    },
    "line_style": {
        "linewidth": 2,
    },
    "scattershadow_style": {
        "s": 30,
    },
    "lineshadow_style": {
        "linewidth": 2,
    },
    "styleParams": {
        "xticks": {"fontsize": 11, "fontweight": "bold"},
        "yticks": {"fontsize": 11, "fontweight": "bold"},
        "ylabel": {"fontsize": 12, "fontweight": "bold"},
        "xlabel": {"fontsize": 12, "fontweight": "bold"},
        "title": {
            "fontsize": 18,
            "color": "white",
            "fontweight": "bold",
            "xy": (0.00, 1.16),
            "xycoords": "axes fraction",
        },
        "subtitle": {
            "color": "white",
            "fontsize": 14,
            "xy": (0.00, 1.095),
            "xycoords": "axes fraction",
        },
    },
    "colors": {"1cat": "#aed7eb", "ncats": plt_tab_colors, "grayed": "#566a6b"},
}



from setuptools import find_packages, setup

base_dependencies = [
    "pandas>=1.0.0",
    "python-dateutil>=2",
    "seaborn>0.9.0",
    "rutil>=0.0.1",
    "matplotlib>=2.2",
]


additional_dependencies = {
    "dev": ["black>=21.9b0", "pre-commit>=2.15.0", "pytest>=6.2.1", "pylint>=2.7.4", "jupyterlab", "twine"],
}

VERSION = "0.0.3"
DESCRIPTION = "A matplotlib/seaborn wrapper to create beautiful plots with predefined styles using pipelines"

with open("README.md", "r", encoding="utf8") as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name="pltflow",
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Ismael Cabral",
    version=VERSION,
    keywords=["matplotlib", "seaborn", "plt", "plot", "graphs", "data", "visualization", "pandas"],
    install_requires=base_dependencies,
    extras_require=additional_dependencies,
    python_requires=">=3.6",
)
